/**************************************************************************/
/* File: write_rtl.c - RT-Linux write systemcall variant                  */
/*                                                                        */
/* LinCAN - (Not only) Linux CAN bus driver                               */
/* Copyright (C) 2002-2009 DCE FEE CTU Prague <http://dce.felk.cvut.cz>   */
/* Copyright (C) 2002-2009 Pavel Pisa <pisa@cmp.felk.cvut.cz>             */
/* Funded by OCERA and FRESCOR IST projects                               */
/* Based on CAN driver code by Arnaud Westenberg <arnaud@wanadoo.nl>      */
/*                                                                        */
/* LinCAN is free software; you can redistribute it and/or modify it      */
/* under terms of the GNU General Public License as published by the      */
/* Free Software Foundation; either version 2, or (at your option) any    */
/* later version.  LinCAN is distributed in the hope that it will be      */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty    */
/* of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU    */
/* General Public License for more details. You should have received a    */
/* copy of the GNU General Public License along with LinCAN; see file     */
/* COPYING. If not, write to the Free Software Foundation, 675 Mass Ave,  */
/* Cambridge, MA 02139, USA.                                              */
/*                                                                        */
/* To allow use of LinCAN in the compact embedded systems firmware        */
/* and RT-executives (RTEMS for example), main authors agree with next    */
/* special exception:                                                     */
/*                                                                        */
/* Including LinCAN header files in a file, instantiating LinCAN generics */
/* or templates, or linking other files with LinCAN objects to produce    */
/* an application image/executable, does not by itself cause the          */
/* resulting application image/executable to be covered by                */
/* the GNU General Public License.                                        */
/* This exception does not however invalidate any other reasons           */
/* why the executable file might be covered by the GNU Public License.    */
/* Publication of enhanced or derived LinCAN files is required although.  */
/**************************************************************************/

#ifdef CAN_WITH_RTL

#include "../include/can.h"
#include "../include/can_sysdep.h"
#include "../include/main.h"

#include <rtl_posixio.h>
#include "../include/can_iortl.h"

ssize_t can_write_rtl_posix(struct rtl_file *fptr, const char *buffer,
				 size_t length, loff_t *ppos)
{
	struct canuser_t *canuser =
		(struct canuser_t *)can_get_rtl_file_private_data(fptr);
	const struct canmsg_t *msg_buff = (struct canmsg_t *)buffer;
	struct canque_ends_t *qends;
	struct canque_edge_t *qedge;
	struct canque_slot_t *slot;
	int      msg_flags;
	int      ret;
	size_t   bytes_to_copy;

	if(!canuser || (canuser->magic != CAN_USER_MAGIC)){
		CANMSG("can_write_rtl_posix: bad canuser magic\n");
		return -ENODEV;
	}

	if (length < sizeof(struct canmsg_t)) {
		DEBUGMSG("Trying to write less bytes than a CAN message,\n");
		DEBUGMSG("this will always return 0 !\n");
		return 0;
	}
	if (length > INT_MAX) {
		CANMSG("Trying to write more than is supported.\n");
		return -1;
	}

	qends = canuser->qends;

	msg_flags=msg_buff->flags;

	/* Automatic selection of extended format if ID>2047 */
	if (msg_buff->id & ~0x7ffl & MSG_ID_MASK ) msg_flags |= MSG_EXT;
	/* has been dependent on "extended" option */

	/* If the output buffer is full, return immediately in case O_NONBLOCK
	 * has been specified or loop until space becomes available.
	 */
	if ((ret=canque_get_inslot4id(qends, &qedge, &slot,
			0, msg_buff->id, 0))<0){
		DEBUGMSG("Buffer is full\n");
		if(ret < -1)
			return -EIO;

		if (fptr->f_flags & O_NONBLOCK)
			return -EAGAIN;

		ret=canque_get_inslot4id_wait_rtl(qends, &qedge, &slot,
						0, msg_buff->id, 0);
		if (ret == -1)
			return -EINTR;

		if(ret<0) {
			return -EIO;
		}
	}
	slot->msg=*(msg_buff++);
	slot->msg.flags=msg_flags;
	canque_put_inslot(qends, qedge, slot);
	bytes_to_copy = length-sizeof(struct canmsg_t);

	/*
	 * Try to send more messages
	 */
	while (bytes_to_copy >= sizeof(struct canmsg_t)) {
		bytes_to_copy -= sizeof(struct canmsg_t);
		msg_flags=msg_buff->flags;

		/* Automatic selection of extended format if ID>2047 */
		if (msg_buff->id & ~0x7ffl & MSG_ID_MASK ) msg_flags |= MSG_EXT;
		/* has been dependent on "extended" option */

		/* Get slot */
		if(canque_get_inslot4id(qends, &qedge, &slot,
			0, msg_buff->id, 0) < 0) break;
		slot->msg=*(msg_buff++);
		slot->msg.flags=msg_flags;
		canque_put_inslot(qends, qedge, slot);
	}

        if(fptr->f_flags & O_SYNC) {
		canque_sync_wait_rtl(qends, qedge);
	}
	return length-bytes_to_copy;
}

#endif /*CAN_WITH_RTL*/
